package ssoexample;

import com.sun.security.auth.callback.TextCallbackHandler;
import org.apache.commons.cli.*;
import org.apache.log4j.Logger;
import org.ietf.jgss.*;
import org.jboss.netty.bootstrap.ClientBootstrap;
import org.jboss.netty.channel.*;
import org.jboss.netty.channel.socket.nio.NioClientSocketChannelFactory;
import org.jboss.netty.handler.codec.serialization.ObjectDecoder;
import org.jboss.netty.handler.codec.serialization.ObjectEncoder;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import java.net.InetSocketAddress;
import java.security.PrivilegedExceptionAction;
import java.util.concurrent.Executors;

public class Client {
    private ChannelFuture channelFuture;
    private static final Logger logger = Logger.getLogger(Client.class);

    public Client(String[] args) throws Exception {
        try {
            Options options = new Options();
            options.addOption("n", true, "server hostname");
            options.addOption("r", true, "realm, e.g. EXAMPLE.COM");
            options.addOption("k", true, "KDC server");
            options.addOption("p", true, "server port");
            options.addOption("s", true, "SPN, e.g. test_sso");
            options.addOption("s", true, "SPN, e.g. <service>");
            options.addOption("h", false, "print this message");

            CommandLine cmd = new PosixParser().parse(options, args);

            if (cmd.hasOption("h") || !(cmd.hasOption("n") && cmd.hasOption("p") && cmd.hasOption("r") && cmd.hasOption("k")) && cmd.hasOption("s")) {
                HelpFormatter formatter = new HelpFormatter();
                formatter.printHelp("client", options);
                System.exit(0);
            }

            String allowtgtsessionkey = RegQuery.query("\"HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Lsa\\Kerberos\\Parameters\"", "allowtgtsessionkey");
            if (allowtgtsessionkey == null || Integer.decode(allowtgtsessionkey) != 1) {
                logger.error("allowtgtsessionkey must be set");
                System.exit(-1);
            }

            System.setProperty("java.security.krb5.kdc", cmd.getOptionValue("k"));
            System.setProperty("java.security.krb5.realm", cmd.getOptionValue("r"));

            Oid krb5Oid = new Oid("1.2.840.113554.1.2.2");
            GSSManager manager = GSSManager.getInstance();

            String spn = cmd.getOptionValue("s");
            String server = cmd.getOptionValue("n");

            String serverPrinc = spn + "@" + server;
            GSSName serverName = manager.createName(serverPrinc, GSSName.NT_HOSTBASED_SERVICE);

            final GSSContext context = manager.createContext(serverName, krb5Oid, null, GSSContext.DEFAULT_LIFETIME);
            context.requestMutualAuth(true);
            context.requestConf(true);
            context.requestInteg(true);

            final SimpleChannelUpstreamHandler handler = new SimpleChannelUpstreamHandler() {
                @Override
                public void channelConnected(ChannelHandlerContext ctx, ChannelStateEvent e) {
                    e.getChannel().write(new HelloMessage());
                }

                @Override
                public void messageReceived(ChannelHandlerContext ctx, MessageEvent event) {
                    logger.info("message received: " + event.getMessage());
                    if (event.getMessage() instanceof AuthorizationRequiredMessage) {
                        try {
                            if (!context.isEstablished()) {
                                byte[] token = new byte[0];
                                try {
                                    token = context.initSecContext(token, 0, token.length);
                                } catch (GSSException e) {
                                    logger.error("error trying to authenticate, local admin user?", e);
                                    System.exit(-1);
                                }
                                sendToken(token);
                            }

                        } catch (Exception e) {
                            throw new RuntimeException(e);
                        }
                    } else if (event.getMessage() instanceof TokenMessage) {
                        try {
                            byte[] token = new byte[0];
                            TokenMessage tokenMessage = (TokenMessage) event.getMessage();
                            if (!context.isEstablished()) {
                                token = context.initSecContext(tokenMessage.getToken(), 0, token.length);
                                sendToken(token);
                            }
                        } catch (Exception e) {
                            throw new RuntimeException(e);
                        }
                    } else if (event.getMessage() instanceof AuthorizedMessage) {
                        logger.info("client authorized with server successfully");

                        channelFuture.getChannel().disconnect();
                        channelFuture.getChannel().unbind();
                        channelFuture.getChannel().close();
                        System.exit(0);
                    }
                }

                private void sendToken(byte[] token) {
                    channelFuture.getChannel().write(new TokenMessage(token));
                    logger.info("sending token to server: " + TokenMessage.byteArrayToHexString(token));
                }

                @Override
                public void exceptionCaught(ChannelHandlerContext ctx, ExceptionEvent e) {
                    logger.error("exception caught in server channel handler", e.getCause());
                }

            };
            NioClientSocketChannelFactory channelFactory = new NioClientSocketChannelFactory(Executors.newCachedThreadPool(), Executors.newCachedThreadPool());
            ClientBootstrap bootstrap = new ClientBootstrap(channelFactory);

            bootstrap.setPipelineFactory(new ChannelPipelineFactory() {
                public ChannelPipeline getPipeline() throws Exception {
                    return Channels.pipeline(
                            new ObjectEncoder(),
                            new ObjectDecoder(),
                            handler);
                }
            });

            channelFuture = bootstrap.connect(new InetSocketAddress(cmd.getOptionValue("n"), Integer.parseInt(cmd.getOptionValue("p"))));
            logger.info("client connected");
        } catch (ParseException e) {
            throw new RuntimeException(e);
        }
    }

    public static void main(final String[] args) {
        try {
            System.setProperty("java.security.auth.login.config", Config.JAAS_CONF_FILENAME);
            System.setProperty("sun.security.krb5.debug", "true");
            System.setProperty("sun.security.jgss.debug", "true");
            System.setProperty("javax.security.auth.useSubjectCredsOnly", "false");

            CallbackHandler callbackHandler = new TextCallbackHandler();

            LoginContext context = null;

            try {
                context = new LoginContext("client", callbackHandler);
                context.login();
            } catch (LoginException e) {
                logger.error("Login failed", e);
                System.exit(-1);
            }

            Subject subject = context.getSubject();
            Subject.doAs(subject, new PrivilegedExceptionAction<Object>() {
                public Object run() throws Exception {
                    return new Client(args);
                }
            });

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
}
