/*
 * Copyright 2012 Entero Corporation. All Rights Reserved.
 * www.entero.com
 */
package ssoexample;

import com.google.common.io.Files;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.apache.log4j.Logger;
import sun.security.krb5.PrincipalName;
import sun.security.krb5.internal.ktab.KeyTab;

import java.io.File;

public class Config {
    public static final String KEYTAB_FILENAME = "test.keytab";
    public static final String KRB5_CONF_FILENAME = "krb5.conf";
    public static final String JAAS_CONF_FILENAME = "jaas.conf";
    private final Logger logger = Logger.getLogger(getClass());

    public Config(String[] args) throws Exception {
        Options options = new Options();
        options.addOption("r", true, "realm, e.g. EXAMPLE.COM");
        options.addOption("k", true, "KDC server");
        options.addOption("p", true, "service account password");
        options.addOption("s", true, "SPN, e.g. <service>/<fully qualified hostname>");
        options.addOption("h", false, "print this message");

        CommandLine cmd = new PosixParser().parse(options, args);
        if (cmd.hasOption("h") || !(cmd.hasOption("s") && cmd.hasOption("p") && cmd.hasOption("r") && cmd.hasOption("k"))) {
            HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("config", options);
            System.exit(0);
        }

        String spn = cmd.getOptionValue("s");
        String password = cmd.getOptionValue("p");

        String realm = cmd.getOptionValue("r");

        createKrb5Conf(realm, cmd.getOptionValue("k"));

        File keyTabFile = new File(KEYTAB_FILENAME);
        if (keyTabFile.exists()) {
            logger.info("deleting previous keytab file " + keyTabFile.getCanonicalPath());
            keyTabFile.delete();
        }
        System.setProperty("java.security.krb5.conf", KRB5_CONF_FILENAME);

        KeyTab keyTab = KeyTab.create(KEYTAB_FILENAME);

        PrincipalName principalname = new PrincipalName(spn);
        if (principalname.getRealm() == null) {
            principalname.setRealm(realm);
        }
        keyTab.addEntry(principalname, password.toCharArray(), 0);
        keyTab.save();
    }

    private void createKrb5Conf(String realm, String kdc) throws Exception {
        if (!realm.toUpperCase().equals(realm)) {
            logger.error("realm should be uppercase");
        }
        StringBuilder sb = new StringBuilder();
        sb.append("[libdefaults]\n");
        sb.append(" default_realm = ").append(realm).append("\n");
        sb.append(" default_tkt_enctypes = aes128-cts rc4-hmac des3-cbc-sha1 des-cbc-md5 des-cbc-crc\n");
        sb.append(" default_tgs_enctypes = aes128-cts rc4-hmac des3-cbc-sha1 des-cbc-md5 des-cbc-crc\n");
        sb.append(" permitted_enctypes   = aes128-cts rc4-hmac des3-cbc-sha1 des-cbc-md5 des-cbc-crc\n\n");
        sb.append("[realms]\n");
        sb.append(realm).append(" = {\n");
        sb.append(" kdc = ").append(kdc).append("\n");
        sb.append("}\n\n");

        sb.append("[domain_realm]\n");
        sb.append(" .").append(realm.toLowerCase()).append(" = ").append(realm).append("\n");
        sb.append(" ").append(realm.toLowerCase()).append(" = ").append(realm).append("\n\n");

        sb.append("[appdefaults]\n");
        sb.append(" kinit = {\n");
        sb.append("  renewable = true\n");
        sb.append("  forwardable = true\n");
        sb.append(" }");

        File krb5ConfFile = new File(KRB5_CONF_FILENAME);
        if (krb5ConfFile.exists()) {
            logger.info("deleting previous krb5conf file " + krb5ConfFile.getCanonicalPath());
            krb5ConfFile.delete();
        }

        Files.write(sb.toString().getBytes(), krb5ConfFile);
    }

    public static void main(String[] args) {
        try {
            new Config(args);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
}
