package ssoexample;

import com.sun.security.auth.callback.TextCallbackHandler;
import org.apache.commons.cli.*;
import org.apache.log4j.Logger;
import org.ietf.jgss.*;
import org.jboss.netty.bootstrap.ServerBootstrap;
import org.jboss.netty.channel.*;
import org.jboss.netty.channel.socket.nio.NioServerSocketChannelFactory;
import org.jboss.netty.handler.codec.serialization.ObjectDecoder;
import org.jboss.netty.handler.codec.serialization.ObjectEncoder;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import java.net.InetSocketAddress;
import java.security.PrivilegedExceptionAction;
import java.util.concurrent.Executors;

public class Server {
    private final Logger logger = Logger.getLogger(getClass());
    private GSSContext context;

    public Server(final String[] args) throws Exception {
        try {
            Options options = new Options();
            options.addOption("n", true, "server hostname");
            options.addOption("p", true, "server port");
            options.addOption("s", true, "SPN, e.g. <service>/<fully qualified hostname>");
            options.addOption("h", false, "print this message");

            final CommandLine cmd = new PosixParser().parse(options, args);

            if (cmd.hasOption("h") || !(cmd.hasOption("n") && cmd.hasOption("p") && cmd.hasOption("s"))) {
                HelpFormatter formatter = new HelpFormatter();
                formatter.printHelp("server", options);
                System.exit(0);
            }

            String spn = cmd.getOptionValue("s");
            logger.info("spn is " + spn);
            System.setProperty("sun.security.krb5.principal", spn);
            System.setProperty("java.security.auth.login.config", Config.JAAS_CONF_FILENAME);
            System.setProperty("java.security.krb5.conf", Config.KRB5_CONF_FILENAME);
            System.setProperty("sun.security.krb5.debug", "true");
            System.setProperty("sun.security.jgss.debug", "true");

            CallbackHandler callbackHandler = new TextCallbackHandler();

            LoginContext context = null;

            try {
                context = new LoginContext("server", callbackHandler);
                context.login();
            } catch (LoginException e) {
                logger.error("Login failed", e);
                System.exit(-1);
            }

            Subject subject = context.getSubject();
            Subject.doAs(subject, new PrivilegedExceptionAction<Object>() {
                public Object run() throws Exception {
                    startServer(cmd);
                    return null;
                }
            });

        } catch (ParseException e) {
            throw new RuntimeException(e);
        }
    }

    private void startServer(CommandLine cmd) throws GSSException {
        final GSSManager manager = GSSManager.getInstance();
        Oid krb5Mechanism = new Oid("1.2.840.113554.1.2.2");
        final GSSCredential serverCreds = manager.createCredential(null, GSSCredential.DEFAULT_LIFETIME, krb5Mechanism, GSSCredential.ACCEPT_ONLY);

        NioServerSocketChannelFactory channelFactory = new NioServerSocketChannelFactory(Executors.newCachedThreadPool(), Executors.newCachedThreadPool());
        ServerBootstrap bootstrap = new ServerBootstrap(channelFactory);

        bootstrap.setPipelineFactory(new ChannelPipelineFactory() {
            public ChannelPipeline getPipeline() throws Exception {
                return Channels.pipeline(
                        new ObjectEncoder(),
                        new ObjectDecoder(),
                        new SimpleChannelUpstreamHandler() {
                            @Override
                            public void messageReceived(ChannelHandlerContext ctx, MessageEvent event) throws Exception {
                                if (!(event.getMessage() instanceof TokenMessage)) {
                                    logger.info("message received: " + event.getMessage());
                                }
                                if (event.getMessage() instanceof HelloMessage) {
                                    context = manager.createContext(serverCreds);
                                    event.getChannel().write(new AuthorizationRequiredMessage());
                                } else if (event.getMessage() instanceof TokenMessage) {
                                    TokenMessage tokenMessage = (TokenMessage) event.getMessage();
                                    logger.info("token recieved from client: length: " + tokenMessage.getToken().length + " content: " + TokenMessage.byteArrayToHexString(tokenMessage.getToken()));
                                    if (!context.isEstablished()) {
                                        byte[] token = tokenMessage.getToken();
                                        token = context.acceptSecContext(token, 0, token.length);
                                        if (!context.isEstablished() && token != null) {
                                            logger.info("sending token to client: length:  " + token.length + " content: " + TokenMessage.byteArrayToHexString(token));
                                            event.getChannel().write(new TokenMessage(token));
                                        }

                                        if (context.isEstablished()) {
                                            logger.info("client principal: " + context.getSrcName());
                                            logger.info("server principal: " + context.getTargName());
                                            if (context.getMutualAuthState()) {
                                                logger.info("mutual authentication");
                                                event.getChannel().write(new AuthorizedMessage());
                                            }
                                        }
                                    }
                                }
                            }

                            @Override
                            public void exceptionCaught(ChannelHandlerContext ctx, ExceptionEvent e) throws Exception {
                                logger.error("exception caught in server channel handler", e.getCause());
                            }
                        });
            }
        });

        bootstrap.bind(new InetSocketAddress(cmd.getOptionValue("n"), Integer.parseInt(cmd.getOptionValue("p"))));
        logger.info("server started");
    }

    public static void main(final String[] args) {
        try {
            new Server(args);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }
}
